#ifndef _BACKUP_H
#define _BACKUP_H

#include <VFSMgr.h>
#include "Backup_Res.h"
#include "sects.h"
#include "zip.h"

typedef enum {
   NOOP,
   BACKUP,
   SAVE_COMMENT
} ActionType;

typedef struct {
    Char*        path;
    Char*        friendly;
    Boolean      incomplete;
    Boolean      automatic;
} SetType;

typedef enum {
    INFO,
    ERROR,
    FATALERROR,
    IMPORTANT,
    FROM_SERVER,
    TO_SERVER,
#ifdef DEBUG
    DEBUGCOLOR
#endif
} MessageType;

void Pause( UInt16 n );
void Message( MessageType m, const Char* s );

#ifdef DEBUG
#define Debug( s ) ( Message( DEBUGCOLOR, s ), Pause( 3 ) )
#else
#define Debug( s )
#endif

#define HIDDEN     0
#define CARD       1

// The following types are internal to Backup.c
#define INTERNAL_DRIVE 2
#define NON_REMOVABLE  3
#define STRICT_SD 10

typedef enum {
    BACKUP_FILES = 0,
    DIRS = 1,
    ALL  = 2
} ScanType;

typedef enum {
    FLUSH_AUTO   = 0,
    FLUSH_NEVER  = 1,
    FLUSH_ALWAYS = 2,
} FlushType;

#define PASSWORD_SCRATCH_SPACE ( 2 + 7 )

typedef struct {
    Boolean  havePassword;
    Boolean  defunct1;
    UInt8    password[ MAX_PASSWORD_SIZE + PASSWORD_SCRATCH_SPACE ];
    Char     host[ MAX_SERVER_SIZE + 1 ];
    Char     user[ MAX_USER_SIZE + 1 ];
    Char     baseDir[ MAX_BASEDIR_SIZE + 1 ];
    Boolean  defunct2;
    UInt16   port;
} ServerSettings;

typedef struct {
    UInt8 scheduleHour_defunct;
    UInt8 scheduleMinute_defunct;
    UInt8 scheduleHalfDay_defunct;
    Boolean daily_defunct;
    UInt8   cycle_defunct;
    Boolean didWarn;
    Int8    compression;
    Boolean slowUpdate;
    Boolean testMode;
    Boolean noBacklight_defunct;
    Boolean resetAfterScheduled;
    UInt32  forceNextAlarm;
    UInt8   scheduleHour;
    UInt8   scheduleMinute;
    Boolean daily;
    UInt8   cycle;
    FlushType flush_defunct;
    FlushType flush;
    UInt8    didLicense:1;
    UInt8    didHelp:1;
    UInt8    didWarnInternalDrive:1;
    Boolean  copySelf;
    Boolean  trustDates;
    Boolean  internalDrive;
    ServerSettings server;
    Boolean  exportAfterBackup;
    Boolean  dimOnTransfer_defunct;
    Boolean  dimOnScheduled;
    Char     devicePassword[ MAX_PASSWORD_SIZE + PASSWORD_SCRATCH_SPACE ];
    Boolean  haveDevicePassword;
    Boolean  needLock;
    Boolean  newCyclingWarned;
    Char     key[ MAX_PASSWORD_SIZE + PASSWORD_SCRATCH_SPACE ];
    Boolean  haveKey;
    Boolean  doCrypt;
    Boolean  setTime;
    Char     timeServer[ MAX_SERVER_SIZE + 1 ];
    Boolean  didCompatibility;
    UInt32   lastScheduledTime;
    Boolean  disableFewer;
} PreferenceType;

extern PreferenceType pref;

#define MAX_FILENAME 256

typedef UInt8 VolumeType;

UInt32 StorageSize( Boolean cache, UInt32* freeP, UInt32* maxP );
Boolean DBCacheFlush( void );

#define THICK -1

Boolean SaveComment( UInt16 vol, Char* dir, Char* comment );
void LoadMessageFont( void );
void ClearScreen( void );
void ProgressBar( Int16 barNumber, UInt32 current, UInt32 total );
void DoRawBackup( Char* cardPath, VolumeType destType, Int8 curCompressLevel );
void* SafeMemPtrNew( UInt32 size );
void SafeMemPtrFree( void* p );
Boolean FindMe( UInt16* card, LocalID* id );
Boolean DoBackup( Char* destDir, VolumeType destType, Char* srcDir, VolumeType srcType,
     Boolean doAll, Boolean partial, Char* comment, Char* renameTo, Int8 compressLevel,
     Char** selectFiles, Int32 numSelectFiles, Boolean* selection);
UInt16 FindVolume( VolumeType type, VolumeInfoType* infoP );
void FatalError( UInt16 alert );
void DeleteUniqueFiles( UInt16 vol, Char* dir, Char** names, Int32 count, Char** compareAgainst, Int32 compareCount );
Boolean MakeDir( UInt16 vol, const Char* dir );
Char** ScanDir( UInt16 vol, const Char* path, Int32* countP, ScanType type );
void FixName( Char* out, Char* in );
void CleanupBackup( void );
void DisableHacks( void );
Char* LastPart( Char* path );
Boolean Handler_frmFTP( EventType* event );
Boolean Confirm( UInt16 alert );
Int16 CompareStrings( void* a, void* b, Int32 discard );
UInt16 GetObjectIndex( UInt16 id );
void* GetObjectPtr( UInt16 id );
Int16 GetSelection( UInt16 lstID );
Char* GetFriendly( Char* lastPart, Boolean forceIncomplete );
Boolean IsIncomplete( Char* s );
Boolean HaveThisDirLocally( Char* dir );
void PasswordDecrypt( Char* dest, UInt8* src );
void PasswordEncrypt( UInt8* dest, Char* src );
FTP_SECTION void Export( UInt16 vol, Char* path );
FTP_SECTION Boolean Handler_frmServer( EventType* event );
FTP_SECTION Boolean Handler_frmFTP( EventType* event );
Boolean ValidBackupName( Char* name );
void Show( UInt16 id, Boolean state );
void SavePrefs( void );
void SetField( UInt16 id, Char* s );
void CopyFromField( Char* dest, UInt16 id, UInt16 bufLen );
void SetValue( UInt16 id, UInt16 value );
UInt16 GetValue( UInt16 id );
Boolean Handler_frmComment( EventType* event );
COMMENT_SECTION Char* PathAppend( Char* path, Char* name );
COMMENT_SECTION void GetComment( Char* path );
COMMENT_SECTION Boolean Handler_frmComment( EventType* event );
void LCDState( Boolean state );
void SetupTrivial( CryptInfo* cP, Char* password );
UInt32 GetCrypt( UInt16 vol, Char* dir );
Char* CryptSetup( CryptInfo* cP, UInt32 cryptID, Char* password );
Boolean KeyVerify( Int16 button, Char* attempt );
FTP_SECTION Boolean TimeServer( Char* hostName, UInt32* newTimeP, Int32* deltaP );
UInt16 SchedulerDialog( void );
Boolean CheckBattery( void );
UInt16 _ProcessButtons( BitmapType* bmp );

extern UInt16  curCardVol;
extern Boolean scheduled;
extern Boolean nvfsOff;
extern Boolean redoList;
extern Char    comment[ MAX_COMMENT_SIZE + 1 ];
extern ActionType afterComment;
extern Int32      afterCommentArgument;
extern ActionType nextAction;
extern Int32      nextActionArgument;
#define DATE_BUFFER_LENGTH ( 4 + 2 + 2  +  2 + 2 + 2 + 1 )

#define LOG_ROLLOVER_SIZE  (512ul*1024)

#define commentName  "%%comment.dat"
#define oldCommentName "%%comment.txt"
#define tempName     "%%temp"
#define logName      "%%backup.log"
#define logName2     "%%backup.log.2"
#define oldLogName   "%%log.txt"
#define oldCryptName    "%%encrypted.txt"
#define cryptName    "%%encrypted.dat"

#define nvBackupPrefix "%%"
#define nvBackupPrefixLength 2

#define backupDir    "/PALM/Programs/NVBackup"
#define incompleteMarker   " (incomplete)"
#define autoMarker         " (cycle)"
// also edit in Backup.rcp if you change this.  Should be 31 chars long
#define badBackupFileName "NVbkInCompLeteBaCkUp7299302992z"

#define MemZero( p, n ) MemSet( p, n, 0 )

#endif
