/* $Id: SchedulerDialog.c,v 1.8 2008/07/31 04:57:41 arpruss Exp $ */
#include <PalmOS.h>
#include "Backup.h"
#include "endianutils.h"

#define DEFAULT_BUTTON  0
#define DELAY_TIME     20
#define NO_BUTTON      0xFFFF


static UInt16 GetButton( Coord x, Coord y )
{

//    WinDisplayToWindowPt( &x, &y );

    y -= 80;

    if ( y < 130 / 2 || 130 / 2 + 24 / 2 <= y )
        return NO_BUTTON;

    if ( x < 7 / 2 )
        return NO_BUTTON;

    if ( x < 61 / 2 + 1 )
        return 0;

    if ( x < 69 / 2 )
        return NO_BUTTON;

    if ( x < 200 / 2 )
        return 1;

    if ( x < 204 / 2 )
        return NO_BUTTON;

    if ( x < 313 / 2 + 1 )
        return 2;

    return NO_BUTTON;
}



Err HedeHandler( SysNotifyParamType* np )
{
    EventType* e;


    e = ( EventType* )( np->notifyDetailsP );
//    { Char s[55]; StrPrintF(s, "%x",e->eType ); WinDrawChars(s,4,SysRandom(0)%100,SysRandom(0)%100); }
    switch ( e->eType ) {
        case ByteSwap16( penDownEvent ):
            FtrSet( appCreator, FTR_PEN_DOWN, 1 );
            FtrSet( appCreator, FTR_PEN_X, ByteSwap16( e->screenX ) );
            FtrSet( appCreator, FTR_PEN_Y, ByteSwap16( e->screenY ) );
//            WinDrawChar('d',ByteSwap16( e->screenX ),ByteSwap16( e->screenY ));
            break;
        case ByteSwap16( penUpEvent ):
            FtrSet( appCreator, FTR_PEN_DOWN, 0 );
            FtrSet( appCreator, FTR_PEN_X, ByteSwap16( e->screenX ) );
            FtrSet( appCreator, FTR_PEN_Y, ByteSwap16( e->screenY ) );
            break;
        default:
            break;
    }

    return errNone;
}



static void RegisterSchedulerDialog( Boolean state )
{
    UInt16 card;
    LocalID id;

    if ( ! FindMe( &card, &id ) )
        return;

    FtrSet( appCreator, FTR_PEN_DOWN, 0 );

    if ( state ) {
        SysNotifyRegister( card, id, 'hede', HedeHandler, -128, NULL );
//        SysNotifyRegister( card, id, appCreator, NULL, 0, NULL );
    }
    else {
        SysNotifyUnregister( card, id, 'hede', -128 );
//        SysNotifyUnregister( card, id, appCreator, 0 );
    }
}


UInt16 _ProcessButtons( BitmapType* bmp )
{
    UInt32        endTime;
    UInt16        curButton;
    Boolean       inButton;
    EventType     event;

    WinDrawBitmap( bmp, 0, 80 );

    endTime  = TimGetTicks() + DELAY_TIME * SysTicksPerSecond();
    curButton = NO_BUTTON;
    inButton = false;

    while ( TimGetTicks() < endTime ) {
        Int32   x;
        Int32   y;
        Int32   down;
        EventType ignore;

        WinDrawBitmap( bmp, 0, 80 );

        EvtGetEvent( &event, SysTicksPerSecond() * 5 );
        if ( event.eType == nilEvent || ( event.eType == keyDownEvent && event.data.keyDown.chr == 0x1618 ) ) {
            SysHandleEvent( &event );
        }

        FtrGet( appCreator, FTR_PEN_DOWN, &down );
        FtrGet( appCreator, FTR_PEN_X, &x );
        FtrGet( appCreator, FTR_PEN_Y, &y );

        if ( ! down ) {
            if ( curButton != NO_BUTTON && inButton )
                return curButton;
            curButton = NO_BUTTON;
        }
        else {
            UInt16 theButton;

            theButton = GetButton( x, y );

            if ( curButton == NO_BUTTON )
                curButton = theButton;

            inButton = ( curButton == theButton );
        }

        SysTaskDelay( 2 * SysTicksPerSecond() / 100 );
    }

    return DEFAULT_BUTTON;
}



UInt16 ProcessButtons( BitmapType* img )
{
    UInt16 value;

    RegisterSchedulerDialog( true );
    
    value = _ProcessButtons( img );
    EvtFlushPenQueue();

    RegisterSchedulerDialog( false );

    return value;
}


UInt16 SchedulerDialog( void )
{
    WinHandle oldWindow;
    RectangleType bounds;
    Err       err;
    WinHandle save;
    MemHandle h;
    UInt16    button;
    
    bounds.topLeft.x = 0;
    bounds.topLeft.y = 80;
    bounds.extent.x  = 160;
    bounds.extent.y  = 80;

    oldWindow = WinGetDrawWindow();
    WinSetDrawWindow( WinGetDisplayWindow() );

    bounds.topLeft.y = 0;
    save = WinSaveBits( &bounds, &err );
    if ( save == NULL || err != errNone ) {
        if ( oldWindow != NULL )
            WinSetDrawWindow( oldWindow );
        return DEFAULT_BUTTON;
    }

    h = DmGetResource( 'Tbmp', imgSchedulerDialog );
    if ( h != NULL ) {
        BitmapType* img;

        img = MemHandleLock( h );

        button = ProcessButtons( img );

        MemHandleUnlock( h );
        DmReleaseResource( h );
    }
    else {
        button = DEFAULT_BUTTON;
    }

    if ( save != NULL ) {
        WinRestoreBits( save, 0, 0 );
        MemPtrFree( save );
    }
    if ( oldWindow != NULL )
        WinSetDrawWindow( oldWindow );

    EvtFlushPenQueue();
    
    return button;
}

