/* $Id: encrypt.c,v 1.9 2006/05/26 23:15:48 arpruss Exp $ */
#include <PalmOS.h>
#include "Backup.h"
#include "AESLib.h"

#define KEY_SIZE 8
#define DIGIT8( c ) ( ( UInt8 )( c - '0' ) )

typedef Boolean (* CryptSetupFunction )( CryptInfo* cP, Char* password );

CRYPT_SECTION static Boolean NoneSetup( CryptInfo* cP, Char* password );
CRYPT_SECTION static Boolean DESSetup( CryptInfo* cP, Char* password );
#ifdef TEA68K
CRYPT_SECTION static Boolean TEASetup( CryptInfo* cP, Char* password );
CRYPT_SECTION static void TEACrypt( UInt32 numBlocks, UInt8* out, UInt8* in, UInt8* key, Boolean encrypt );
#endif
CRYPT_SECTION static Boolean AESSetup( CryptInfo* cP, Char* password );
#ifdef TRIVIAL
CRYPT_SECTION static Boolean TrivialSetup( CryptInfo* cP, Char* password );
CRYPT_SECTION static void TrivialCrypt( UInt32 numBlocks, UInt8* out, UInt8* in, UInt8* key, Boolean encrypt );
#endif
CRYPT_SECTION static void AESClose( void );

static UInt16 aesRefNum;
static aes_ctx* encCTX = NULL;
static aes_ctx* decCTX = NULL;
static UInt8 zeroVector16[ 16 ] = { 0,0,0,0, 0,0,0,0, 0,0,0,0, 0,0,0,0 };

typedef struct {
    CryptSetupFunction Setup;
    UInt32             id;
    Char*              name;
} CryptMethod;

CryptMethod methods[] = {
    { NoneSetup, 0, "none" },
    { DESSetup, 'DES_', "DES (56-bit)" },
    { AESSetup, 'AES1', "AES (128-bit)" },
#ifdef TRIVIAL
    { TrivialSetup, 'Triv', "Trivial (128-bit)" },
#endif
#ifdef TEA68K
    { TEASetup, 'TEAn', "TEA new (128-bit)" }
#endif
};

#define NUM_METHODS ( sizeof methods / sizeof *methods )


static UInt8 ToByte( Char a, Char b, Char c )
{
     return DIGIT8( a ) * 100 + DIGIT8( b ) * 10 + DIGIT8( c );
}

static void MakeKey( UInt8* key )
{
    UInt16 i;
    UInt16 serialLen;
    UInt8* serial;
    UInt8  baseKey[ 8 ];
    Char   miniString[ 3 ];

    baseKey[ 0 ] = 26;
    baseKey[ 1 ] = 226;
    baseKey[ 2 ] = 244;
    baseKey[ 3 ] = 86;
    baseKey[ 4 ] = 75;
    baseKey[ 5 ] = 133;
    baseKey[ 6 ] = 19;
    baseKey[ 7 ] = 119;
    
    miniString[ 0 ] ='4';
    miniString[ 1 ] ='2';
    miniString[ 2 ] ='2';

    if ( errNone != SysGetROMToken( 0, sysROMTokenSnum, &serial, &serialLen ) ||
         serialLen == 0 || serial[ 0 ] == 0xFF ) {
        serialLen = 3;
        serial    = miniString;
    }

    for ( i = 0 ; i < KEY_SIZE ; i++ )
         key[ i ] = baseKey[ i ] + ToByte( serial[ i % serialLen ], serial[ ( i + 1 ) % serialLen ],
                        serial[ ( i + 2 ) % serialLen ] );
}


// Output buffer must be of length PASSWORD_SCRATCH_SPACE + StrLen( src )
void PasswordEncrypt( UInt8* dest, Char* src )
{
    UInt16 i;
    UInt16 len;
    UInt8  key[ 8 ];

    MakeKey( key );

    len = StrLen( src );

    *dest++ = len / 256;
    *dest++ = len % 256;

    for ( i = 0 ; i < len ; i += 8 ) {
         EncDES( src + i, key, dest + i, true );
    }
}



// Output buffer must be of length PASSWORD_SCRATCH_SPACE + StrLen( src )
void PasswordDecrypt( Char* dest, UInt8* src )
{
    UInt16 length;
    UInt16 i;
    UInt8  key[ 8 ];

    MakeKey( key );

    length = src[ 0 ] * 256 + src[ 1 ];
    src += 2;

    for ( i = 0 ; i < length ; i += 8 ) {
         EncDES( src + i, key, dest + i, false );
    }

    dest[ i ] = 0;
}



static Boolean NoneSetup( CryptInfo* cP, Char* password )
{
    cP->Crypt = NULL;
    cP->Close = NULL;

    return true;
}



/* It's possible to have out == in, but no other overlap is permitted. */
static void DESCrypt( UInt32 numBlocks, UInt8* out, UInt8* in, UInt8* key, Boolean encrypt )
{
    UInt32 i;

    for ( i = 0 ; i < numBlocks ; i++ ) {
         if ( out == in ) {
             UInt8 dataCopy[ 8 ];

             MemMove( dataCopy, out + i * 8, 8 );
             EncDES( dataCopy, key, out + i * 8, encrypt );
         }
         else {
             EncDES( in + i * 8, key, out + i * 8, encrypt );
         }
    }
}



static Boolean OddParity( UInt8 x )
{
    return   ( 1 &
                  ( x ^ ( x >> 1 ) ^ ( x >> 2 ) ^ ( x >> 3 ) ^ ( x >> 4 ) ^
                    ( x >> 5 ) ^ ( x >> 6 ) ^ ( x >> 7 ) )
             );
}


static Boolean DESSetup( CryptInfo* cP, Char* password )
{
    UInt16 i;

    cP->Crypt     = DESCrypt;
    cP->Close     = NULL;
    cP->blockSize = 8;
    cP->keySize   = 8;

    for ( i = 0 ; i < 8 && password[ i ] ; i++ ) {
         if ( OddParity( password[ i ] ) )
             cP->key[ i ] = password[ i ];
         else
             cP->key[ i ] = ( UInt8 )password[ i ] ^ 0x80;
    }

    if ( i == 8 ) {
        for ( ; password[ i ] ; i++ ) {
            cP->key[ i % 8 ] += password[ i ];
        }
    }
    else {
        for ( ; i < 8 ; i++ ) {
             cP->key[ i ] = 0;
        }
    }

    return true;
}



static void AESClose( void )
{
    if ( decCTX == NULL )
        return;

    SafeMemPtrFree( decCTX );
    decCTX = NULL;
    SafeMemPtrFree( encCTX );
    encCTX = NULL;
    AESLib_CloseLibrary( aesRefNum );
}

static void AESCrypt( UInt32 numBlocks, UInt8* out, UInt8* in, UInt8* key, Boolean encrypt )
{
    UInt8 copy[ 16 ];

    if ( 1 || numBlocks == 1 || out == in ) {
        UInt32 i;

        for ( i = 0 ; i < numBlocks ; i++ ) {
            UInt8* src;

            if ( out == in ) {
                MemMove( copy, out + i * 16, 16 );
                src = copy;
            }
            else {
                src = in + i * 16;
            }

            if ( encrypt )
                AESLibEncBlk( aesRefNum, src, out + i * 16, encCTX );
            else
                AESLibDecBlk( aesRefNum, src, out + i * 16, decCTX );
        }
    }
    else {
        UInt32 blen;
        blen = numBlocks * 16;
        if ( encrypt )
            AESLibEncBigBlk( aesRefNum, in, out, &blen, false, zeroVector16, encCTX );
        else
            AESLibDecBigBlk( aesRefNum, in, out, &blen, false, zeroVector16, decCTX );
    }
}



static void FillKey( UInt8* key, UInt16 keySize, Char* password, Boolean wrap )
{
    UInt16 i;

    for ( i = 0 ; i < keySize && password[ i ] ; i++ ) {
         key[ i ] = password[ i ];
    }

    if ( wrap && i < keySize ) {
        UInt16 inLength;
        
        inLength = StrLen( password );
        
        if ( inLength == 0 )
            inLength = 1;

        for ( ; i < keySize && password[ i ] ; i++ ) {
             key[ i ] = password[ i % inLength ];
        }
    }

    for ( ; password[ i ] ; i++ ) {
         key[ i % keySize ] += password[ i ];
    }

}



static Boolean AESSetup( CryptInfo* cP, Char* password )
{
    Err    err;

    if ( decCTX != NULL ) {
        AESClose();
    }

    decCTX = MemPtrNew( sizeof( *decCTX ) );
    if ( decCTX == NULL )
        return false;
    MemZero( decCTX, sizeof( *decCTX ) );

    encCTX = MemPtrNew( sizeof( *encCTX ) );
    if ( encCTX == NULL ) {
        MemPtrFree( decCTX );
        decCTX = NULL;
        return false;
    }
    MemZero( decCTX, sizeof( *decCTX ) );

    err = AESLib_OpenLibrary( &aesRefNum );

    if ( err != errNone ) {
        MemPtrFree( decCTX );
        decCTX = NULL;
        MemPtrFree( encCTX );
        encCTX = NULL;
        return false;
    }

    cP->Crypt     = AESCrypt;
    cP->Close     = AESClose;
    cP->blockSize = 16;
    cP->keySize   = 16;
    
    FillKey( cP->key, 16, password, false );

    AESLibEncKey( aesRefNum, cP->key, 16, encCTX );
    AESLibDecKey( aesRefNum, cP->key, 16, decCTX );

    return true;
}



#ifdef TEA68K
void encipher(const UInt32 *const v, UInt32 *const w,
   const UInt32  * const k)
{
   register UInt32 y=v[0],z=v[1],sum=0,delta=0x9E3779B9ul,n=32;

   while(n-->0)
      {
      y += (z << 4 ^ z >> 5) + z ^ sum + k[sum&3];
      sum += delta;
      z += (y << 4 ^ y >> 5) + y ^ sum + k[sum>>11 & 3];
      }

   w[0]=y; w[1]=z;
}

void decipher(const UInt32 *const v,UInt32 *const w,
   const UInt32 * const k)
{
   register UInt32 y=v[0],z=v[1],sum=0xC6EF3720ul, delta=0x9E3779B9ul,n=32;

   /* sum = delta<<5, in general sum = delta * n */

   while(n-->0)
      {
      z -= (y << 4 ^ y >> 5) + y ^ sum + k[sum>>11 & 3];
      sum -= delta;
      y -= (z << 4 ^ z >> 5) + z ^ sum + k[sum&3];
      }

   w[0]=y; w[1]=z;
}


static void TEACrypt( UInt32 numBlocks, UInt8* out, UInt8* in, UInt8* key, Boolean encrypt )
{
    UInt32 i;

    if ( encrypt ) {
        for ( i = 0 ; i < numBlocks ; i++ ) {
             if ( in == out ) {
                 UInt8 copy[ 8 ];
                 MemMove( copy, in + i * 8, 8 );
                 encipher( ( void* )copy, ( void* )( out + i * 8 ), ( void* )key );
             }
             else {
                 encipher( ( void* )( in + i * 8 ), ( void* )( out + i * 8 ), ( void* )key );
             }
        }
    }
    else {
        for ( i = 0 ; i < numBlocks ; i++ ) {
             if ( in == out ) {
                 UInt8 copy[ 8 ];
                 MemMove( copy, in + i * 8, 8 );
                 decipher( ( void* )copy, ( void* )( out + i * 8 ), ( void* )key );
             }
             else {
                 decipher( ( void* )( in + i * 8 ), ( void* )( out + i * 8 ), ( void* )key );
             }
        }
    }
}




static Boolean TEASetup( CryptInfo* cP, Char* password )
{
    cP->Crypt = TEACrypt;
    cP->Close = NULL;
    cP->blockSize = 8;
    cP->keySize   = 16;

    FillKey( cP->key, 16, password, false );

    return true;
}
#endif


#ifdef TRIVIAL
static void TrivialCrypt( UInt32 numBlocks, UInt8* out, UInt8* in, UInt8* key, Boolean encrypt )
{
    UInt32 i;

    MemMove( out, in, numBlocks * 16 );
    return;

    for ( i = 0 ; i < numBlocks ; i++ ) {
         UInt16 j;
         UInt8* src;
         UInt8* dest;

         src  = in + 16 * i;
         dest = out + 16 * i;

         if ( encrypt ) {
             for ( j = 0 ; j < 16 ; j++ ) {
                  dest[ j ] = src[ j ] + key[ j ];
             }
         }
         else {
             for ( j = 0 ; j < 16 ; j++ ) {
                  dest[ j ] = src[ j ] - key[ j ];
             }
         }
    }
}







static Boolean TrivialSetup( CryptInfo* cP, Char* password )
{
    cP->Crypt = TrivialCrypt;
    cP->Close = NULL;
    cP->blockSize = 16;
    cP->keySize   = 16;

    FillKey( cP->key, 16, password, true );

    return true;
}
#endif


Char* CryptSetup( CryptInfo* cP, UInt32 cryptID, Char* password )
{
    UInt16 i;

    for ( i = 0 ; i < NUM_METHODS ; i++ ) {
         if ( cryptID == methods[ i ].id ) {
             Boolean success;
             success = methods[ i ].Setup( cP, password );
             return success ? methods[ i ].name : NULL;
         }
    }

    return NULL;
}

