/* $Id: time.c,v 1.5 2008/06/21 17:45:18 arpruss Exp $ */
#include <PalmOS.h>
#include <NetMgr.h>
#include <VFSMgr.h>
#include <Unix/sys_socket.h>
#include "Backup.h"
#include "ftp.h"

#define PALM_TO_1900_SECONDS ( -126144000l )

#ifndef SHUT_RDWR
# define SHUT_RDWR netSocketDirBoth
#endif

#if 0
UInt32 Get1900Time( void )
{
    Int16  zone;
    Int16  daylight;

    zone     = PrefGetPreference( prefTimeZone );
    daylight = PrefGetPreference( prefDaylightSavingAdjustment );

    return TimTimeZoneToUTC( TimGetSeconds(), zone, daylight ) - PALM_TO_1900_SECONDS;
}
#endif



/* returns Palm time */
UInt32 Set1900Time( UInt32 time, UInt32* oldTimeP )
{
    Int16  zone;
    Int16  daylight;
    UInt32 palmTime;
    UInt32 oldTime;

    zone     = PrefGetPreference( prefTimeZone );
    daylight = PrefGetPreference( prefDaylightSavingAdjustment );

    palmTime = TimUTCToTimeZone( time, zone, daylight ) + PALM_TO_1900_SECONDS;

    oldTime = TimGetSeconds();

    if ( oldTimeP != NULL )
        *oldTimeP = oldTime;

    if ( palmTime != oldTime )
        TimSetSeconds( palmTime );

    return palmTime;
}



#define SNTP_PORT 123

struct {
    unsigned char livnmode, stratum, poll, prec;
    unsigned long delay, disp, refid;
    unsigned long reftime[2], origtime[2], rcvtime[2], txtime[2];
} pkt;

void prfield(Char *s);

char answer[64];


Boolean SNTPServer(Char* hostname, UInt32* newTimeP, Int32* deltaP )
{
    int fd = -1, len, n;
    struct sockaddr_in saddr, myaddr;
    struct hostent *hp;
    Boolean success;
    UInt32 oldTime;
    UInt32 newTime;

    success = false;

    n = StrLen( hostname );

    /* Resolve server name */

    MemSet(&saddr, sizeof saddr,0);
    saddr.sin_family = AF_INET;
    saddr.sin_port = htons(SNTP_PORT);
    if ((saddr.sin_addr.s_addr = inet_addr(hostname)) == -1) {
	prfield("resolving server name...");
	if ((hp = gethostbyname(hostname)) == NULL) {
	    StrCopy(answer, "resolve error");
	    goto done;
	}
	bcopy(hp->h_addr, &saddr.sin_addr, sizeof saddr.sin_addr);
    }

    /* Get a socket */

    if ((fd = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
	StrCopy(answer, "socket error");
	goto done;
    }

    /* Bind it to some port */

    MemSet(&myaddr, sizeof myaddr,0);
    myaddr.sin_family = AF_INET;
    myaddr.sin_port = htons(3030);
    myaddr.sin_addr.s_addr = INADDR_ANY;

    if (bind(fd, (struct sockaddr *) &myaddr, sizeof myaddr) < 0) {
	StrCopy(answer, "bind error");
	goto done;
    }

    Linger( fd, false, 0 );

    /* Fill in the request packet */

    MemSet(&pkt, sizeof pkt,0);
    pkt.livnmode = 0x0b;	/* version 1, mode 3 (client) */

    /* Send packet to server */

    if ((n = sendto(fd, &pkt, sizeof pkt, 0, (struct sockaddr *) &saddr, sizeof saddr)) != sizeof pkt) {
	StrPrintF(answer, "sendto %d err %d", n, errno);
	goto done;
    }
    prfield("waiting for reply...");

    /* Read reply packet from server */

    len = sizeof pkt;
    if ((n = recvfrom(fd, &pkt, sizeof pkt, 0, (struct sockaddr *) &saddr, &len)) != sizeof pkt) {
	if (errno == netErrTimeout)
	    StrCopy(answer, "no reply from server");
	else
	    StrPrintF(answer, "recvfrom %d err %d", n, errno);
	goto done;
    }

    /* Sanity checks */

    if ((pkt.livnmode & 0xc0) == 0xc0 || !pkt.stratum || !pkt.txtime[0]) {
	StrCopy(answer, "server not synchronized");
	goto done;
    }

    /*
     * NTP timestamps are represented as a 64-bit unsigned fixed-
     * point number, in seconds relative to 0h on 1 January 1900. The integer
     * part is in the first 32 bits and the fraction part in the last 32 bits.
     */

    newTime = Set1900Time( ntohl(pkt.txtime[0]), &oldTime );

    if ( newTimeP != NULL )
        *newTimeP = newTime;

    if ( deltaP != NULL )
        *deltaP = newTime - oldTime;

    success = true;

    StrPrintF(answer, "stratum %d offset %ld", (int) pkt.stratum, newTime-oldTime);

 done:
    if (fd >= 0)
	close(fd);
    prfield(answer);

    return success;
}

void prfield(char *s)
{
//    WinDrawChars( s, StrLen(s), 0,0);
}

/*
copyright 1998
the regents of the university of michigan
all rights reserved

permission is granted to use, copy, create derivative works 
and redistribute this software and such derivative works
for any purpose, so long as the name of the university of 
michigan is not used in any advertising or publicity
pertaining to the use or distribution of this software
without specific, written prior authorization.  if the
above copyright notice or any other identification of the
university of michigan is included in any copy of any
portion of this software, then the disclaimer below must
also be included.

this software is provided as is, without representation
from the university of michigan as to its fitness for any
purpose, and without warranty by the university of
michigan of any kind, either express or implied, including
without limitation the implied warranties of
merchantability and fitness for a particular purpose. the
regents of the university of michigan shall not be liable
for any damages, including special, indirect, incidental, or
consequential damages, with respect to any claim arising
out of or in connection with the use of the software, even
if it has been or is hereafter advised of the possibility of
such damages.
*/




FTP_SECTION Boolean TimeServer( Char* hostName0, UInt32* newTimeP, Int32* deltaP )
{
    NetSocketRef controlSocket;
    UInt32       newTime;
    UInt32       oldTime;
    UInt32       time;
    Char         hostName[ MAX_SERVER_SIZE + 1 ];
    Char*        slash;


    if ( ! netOpen )
        return false;

    StrNCopy( hostName, hostName0, MAX_SERVER_SIZE + 1 );

    slash = StrChr( hostName, '/' );

    if ( slash != NULL )
        *slash = 0;

    if ( slash == NULL ||
         ( StrCaselessCompare( slash + 1, "time" ) &&
           StrCompare( slash  + 1, "37" ) ) ) {
         return SNTPServer( hostName, newTimeP, deltaP );
    }

    controlSocket = NetUTCPOpen( hostName, 0, 37 );

    if ( controlSocket < 0 )
        return false;

    Linger( controlSocket, false, 0 );

    if ( NetUReadN( controlSocket, (void*)&time, 4 ) < 4 ) {
        close( controlSocket );
        return false;
    }

    close( controlSocket );
    controlSocket = -1;

    newTime = Set1900Time( time, &oldTime );

    if ( newTimeP != NULL )
        *newTimeP = newTime;

    if ( deltaP != NULL )
        *deltaP = newTime - oldTime;

    return true;
}
