/* $Id: transfer.c,v 1.6 2007/08/30 04:50:42 arpruss Exp $ */
// to do: When transferring incomplete backup, mark it as such?
#include <PalmOS.h>
#include "Backup.h"
#include "ftp.h"

#define FTP_RETRIES  2
#define MAX_ERRORS  10
#define FILE_LIST_ID 3434

static Char workingDir[ MAX_FILENAME + 1 ];
static Char pathBuf[ MAX_FILENAME + 64 ];
static Char pathBuf2[ MAX_FILENAME + 64 ];
static Char password[ MAX_PASSWORD_SIZE + PASSWORD_SCRATCH_SPACE ];
static SetType* serverSets  = NULL;
static Char**   serverFriendlyNames = NULL;
static Int32    serverSetCount = 0;

FTP_SECTION static void ServerClearListData( void );
FTP_SECTION static Boolean ServerOpen( void );
FTP_SECTION static void ServerDoList( void );
FTP_SECTION static void ServerDrawList( void );
FTP_SECTION static void UpdateFTPDisplay( void );
FTP_SECTION static void SaveFields( void );

static Boolean ServerOpen( void )
{
     LoadMessageFont();

     if ( pref.server.havePassword ) {
         PasswordDecrypt( password, pref.server.password );
         if ( ! FTPOpen( pref.server.host, pref.server.user, password, pref.server.port ) )
             return false;
     }
     else {
         if ( ! FTPOpen( pref.server.host, pref.server.user, NULL, pref.server.port ) )
             return false;
     }

     if ( FTPChDir( pref.server.baseDir ) ) {
         FTPMkDir( "NVBackup" );
         if ( FTPChDir( "NVBackup" ) )
             return true;
     }

     FTPClose();
     return false;
}


void Export( UInt16 cardVol, Char* path )
{
     Int32  count;
     UInt32 i;
     Char** names;
     UInt16 retry;
     Boolean didOpen;
     UInt16  numErrors;
//     UInt16  oldBrightness = 0;

     didOpen      = false;
     numErrors    = 0;

     if ( scheduled && ! pref.server.havePassword )
         return;

     if ( ! *pref.server.host || ! *pref.server.user ) {
         if ( ! scheduled  &&  0 == FrmAlert( altNoServerSetup ) )
             FrmPopupForm( frmFTP );
         return;
     }

     names = ScanDir( cardVol, path, &count, ALL );

     StrPrintF( pathBuf, "%s", LastPart( path ) );

     if ( names != NULL ) {
         for ( retry = 0 ; retry <= FTP_RETRIES ; retry++ ) {
              if ( NetOpen() && ServerOpen() ) {
                  FTPMkDir( pathBuf );

                  if ( FTPChDir( pathBuf ) && FTPPWD( workingDir, MAX_FILENAME ) ) {
                      didOpen = true;
                      break;
                  }
              }
              FTPClose();
              NetClose();
         }

         if ( retry <= FTP_RETRIES ) {
              Int32 commentIndex;

              commentIndex = -1;

#ifdef DIM
              if ( scheduled && pref.dimOnScheduled ) {
                  oldBrightness = SysLCDBrightness( false, 0 );
                  SysLCDBrightness( true, 0 );
              }
#endif              

              for ( i = 0 ; i <= count ; i++ ) {
                   Int32 index;

                   ProgressBar( 0, i, count + 1 );

                   // Do "%%comment.txt" last
                   if ( i < count ) {
                       if ( ! StrCompare( names[ i ], commentName ) ) {
                           commentIndex = i;
                           continue;
                       }
                       else
                           index = i;
                   }
                   else {
                       if ( 0 <= commentIndex )
                           index = commentIndex;
                       else
                           continue;
                   }

                   StrPrintF( pathBuf2, "%s/%s", path, names[ index ] );

                   Message( INFO, names[ index ] );

                   if ( ! FTPSendFromVFS( names[ index ], cardVol, pathBuf2 ) ) {
                       UInt16 retry;

                       didOpen = false;

                       Message( ERROR, "\b...error" );

                       for ( retry = 1 ; retry <= FTP_RETRIES ; retry++ ) {
                           FTPClose();
                           NetClose();

                           SysTaskDelay( SysTicksPerSecond() * 2 );

                           didOpen = false;

                           Message( INFO, "(Retry)" );

                           if ( NetOpen() && FTPOpen( pref.server.host, pref.server.user,
                                                      pref.server.havePassword ? password : NULL,
                                                      pref.server.port ) &&
                                FTPChDir( workingDir ) ) {
                               didOpen = true;
                               if ( FTPSendFromVFS( names[ index ], cardVol, pathBuf2 ) )
                                   break;
                           }
                       }

                       if ( FTP_RETRIES < retry )
                           numErrors++;

                       if ( ! didOpen || MAX_ERRORS <= numErrors )
                           break;
                   }
             }

#ifdef DIM
             if ( scheduled && pref.dimOnScheduled )
                 SysLCDBrightness( true, oldBrightness );
#endif

             FTPClose();
             if ( scheduled )
                 NetCloseImmediate();
             else
                 NetClose();

             if ( 0 < numErrors && didOpen ) {
                 if ( ! scheduled ) {
                     FrmAlert( ( numErrors < MAX_ERRORS ) ? altExportFail :
                         altExportTooManyErrors );
                 }
             }
         }
     }

     if ( ! didOpen && ! scheduled ) {
         FrmAlert( altNetOpenFail );
     }

     for ( i = 0 ; i < count ; i++ )
          SafeMemPtrFree( names[ i ] );
     SafeMemPtrFree( names );

     MemSet( password, MAX_PASSWORD_SIZE + PASSWORD_SCRATCH_SPACE, 0 );
}



Char* ChooseServerSet( void )
{
    Char**    names;
    UInt32    count;
    Int16     selected;
    UInt32    i;
    Err       err;
    Char*     answer;

    names = FTPDirList( &count, true );

    answer   = NULL;
    selected = -1;

    if ( names != NULL ) {
        FormType* newForm;
        FormType* oldForm;

        oldForm = FrmGetActiveForm();

        err = LstNewList( ( void** )&newForm, FILE_LIST_ID, 0, 20, 120,
                     120, stdFont, 120 / FntLineHeight(), 0 );

        if ( err == errNone ) {
            UInt16    listIndex;
            ListType* list;

            listIndex = FrmGetObjectIndex( newForm, FILE_LIST_ID );

            list      = FrmGetObjectPtr( newForm, listIndex );

            LstSetListChoices( list, names, count );
            LstSetSelection( list, -1 );

            selected = LstPopupList( list );
            
            FrmEraseForm( newForm );
            FrmDeleteForm( newForm );
            FrmSetActiveForm( oldForm );
        }

        for ( i = 0 ; i < count ; i++ ) {
             if ( 0 <= selected && i == selected ) {
                 answer = names[ i ];
                 continue;
             }
             SafeMemPtrFree( names[ i ] );
        }
        SafeMemPtrFree( names );
    }

    return answer;
}



// check for incompletes
static void ServerDoList( void )
{
    Char** dirs;

    ServerClearListData();

    if ( ! FTPChDir( workingDir ) )
        return;

    dirs = FTPDirList( &serverSetCount, true );

    if ( 0 == serverSetCount ) {
        serverSets = NULL;
    }
    else {
        Int32 i;
        Int32 j;

        SysInsertionSort( dirs, serverSetCount, sizeof( Char* ), CompareStrings, 0 );

        serverSets = SafeMemPtrNew( sizeof( *serverSets ) * serverSetCount );
        serverFriendlyNames = SafeMemPtrNew( sizeof( *serverFriendlyNames ) * serverSetCount );

        j = 0;

        for ( i = 0 ; i < serverSetCount ; i++ ) {
             Boolean forceIncomplete;
             Boolean skip;

             skip = false;
             forceIncomplete = false;

             if ( FTPChDir( dirs[ i ] ) ) {
                 forceIncomplete = ! FTPFetchString( commentName, MAX_COMMENT_SIZE + 1, comment );
                 if ( forceIncomplete ) {
                     /* do one retry just in case */
                     forceIncomplete = ! FTPFetchString( commentName, MAX_COMMENT_SIZE + 1, comment );
                 }
             }
             else {
                 skip = true;
             }

             if ( ! FTPChDir( workingDir ) ) {
                 FTPClose();
                 NetClose();
                 if ( ! NetOpen() || ! ServerOpen() ) {
                     Int32 k;

                     for ( k = 0 ; k < j ; k++ )
                          SafeMemPtrFree( serverFriendlyNames[ k ] );

                     for ( k = 0 ; k < serverSetCount ; k++ )
                          SafeMemPtrFree( dirs[ k ] );

                     SafeMemPtrFree( serverFriendlyNames );
                     SafeMemPtrFree( dirs );
                     serverSetCount = 0;
                     FrmAlert( altNetOpenFail );
                     return;
                 }
             }

             if ( skip )
                 continue;

             serverSets[ j ].path = dirs[ i ];

             serverSets[ j ].friendly = GetFriendly( dirs[ i ], forceIncomplete );

             serverSets[ j ].incomplete = forceIncomplete || IsIncomplete( dirs[ i ] );

             serverFriendlyNames[ j ] = serverSets[ i ].friendly;

             j++;
        }

        serverSetCount = j;

        SafeMemPtrFree( dirs );
    }
}


static void ServerDrawList( void )
{
    LstSetListChoices( GetObjectPtr( lstSets ), serverFriendlyNames, serverSetCount );
    LstSetSelection( GetObjectPtr( lstSets ), -1 );
    LstDrawList( GetObjectPtr( lstSets ) );
}



static void DrawServerForm( void )
{
    FrmDrawForm( FrmGetActiveForm() );
    ServerDrawList();
}



static void ServerClearListData( void )
{
    UInt16 i;

    for ( i = 0 ; i < serverSetCount ; i++ ) {
         SafeMemPtrFree( serverSets[ i ].path );
         SafeMemPtrFree( serverSets[ i ].friendly );
    }
    SafeMemPtrFree( serverSets );
    serverSets = NULL;
    serverSetCount = 0;
    SafeMemPtrFree( serverFriendlyNames );
    serverFriendlyNames = NULL;
}



static void ServerDelete( Char* dir )
{
    Char** files;
    Int32  count;
    Int32  i;

    if ( ! FTPChDir( workingDir ) || ! FTPChDir( dir ) ) {
        FTPClose();
        NetClose();
        NetOpen();
        if ( ! ServerOpen() || ! FTPChDir( workingDir ) || ! FTPChDir( dir ) ) {
            FrmAlert( altNetOpenFail );
            return;
        }
    }

    files = FTPDirList( &count, false );

    for ( i = 0 ; i < count ; i++ ) {
         ProgressBar( -1, i, count );
         if ( StrCompare( files[ i ], "." ) && StrCompare( files[ i ], ".." ) ) {
             FTPDelete( files[ i ] );
             MemPtrFree( files[ i ] );
         }
    }

    ProgressBar( -1, 1, 1 );

    SafeMemPtrFree( files );

    if ( ! FTPChDir( workingDir ) || ! FTPRmDir( dir ) ) {
        FrmAlert( altServerError );
    }

    ServerDoList();
    DrawServerForm();
}



static void ServerImport( Char* dir, Boolean isIncomplete )
{
    Char** files;
    Int32  count;
    Int32  i;
    Boolean success;

    if ( HaveThisDirLocally( dir ) ) {
        FrmAlert( altHaveThisPath );
        return;
    }

    StrPrintF( pathBuf, "%s/%s%s", backupDir, dir, incompleteMarker );

    if ( errNone != VFSDirCreate( curCardVol, pathBuf ) ) {
        FrmAlert( altErrorWriting );
        return;
    }

    if ( ! FTPChDir( workingDir ) || ! FTPChDir( dir ) ) {
        FTPClose();
        NetClose();
        NetOpen();
        if ( ! ServerOpen() || ! FTPChDir( workingDir ) || ! FTPChDir( dir ) ) {
            FrmAlert( altNetOpenFail );
            return;
        }
    }

    files = FTPDirList( &count, false );

    success = 0 < count;

    for ( i = 0 ; i < count ; i++ ) {
         ProgressBar( -1, i, count );

         if ( ValidBackupName( files[ i ] ) ||
              ! StrNCompare( files[ i ], nvBackupPrefix, nvBackupPrefixLength ) ) {
             Message( INFO, files[ i ] );

             StrPrintF( pathBuf2, "%s/%s", pathBuf, files[ i ] );

             if ( ! FTPFetchToVFS( files[ i ], curCardVol, pathBuf2 ) ) {
                 Message( ERROR, "Error importing file" );
                 SysTaskDelay( SysTicksPerSecond() * 1 );
                 success = false;
             }
         }

         MemPtrFree( files[ i ] );
    }

    ProgressBar( -1, 1, 1 );

    SafeMemPtrFree( files );

    redoList = true;

    if ( success ) {
        if ( ! isIncomplete ) {
            VFSFileRename( curCardVol, pathBuf,  dir );
        }
    }
    else {
        FrmAlert( altErrorInImport );
    }
}



Boolean Handler_frmServer( EventType* event )
{
    Boolean handled;

    handled = false;

    switch (event->eType) {
        case frmOpenEvent:
            if ( NetOpen() &&
                 ServerOpen() &&
                 FTPPWD( workingDir, MAX_FILENAME ) ) {
                ServerDoList();
                DrawServerForm();
            }
            else {
                FrmAlert( altNetOpenFail );
                FTPClose();
                NetClose();
                FrmGotoForm( frmMain );
            }
            handled = true;
            break;

        case frmUpdateEvent:
            DrawServerForm();
            handled = true;
            break;

        case frmCloseEvent:
            ServerClearListData();
            FTPClose();
            NetClose();
            handled = true;
            break;

        case ctlSelectEvent: {
            UInt16       id;
            Int16        sel;

            sel = GetSelection( lstSets );

            id = event->data.ctlEnter.controlID;

            switch( event->data.ctlEnter.controlID ) {
                case btnImport:
                    if ( 0 <= sel )
                        ServerImport( serverSets[ sel ].path, serverSets[ sel ].incomplete );
                    else
                        FrmAlert( altSelectSet );
                    DrawServerForm();

                    handled = true;
                    break;
                case btnDisconnect:
                    FTPClose();
                    NetClose();
                    FrmGotoForm( frmMain );
                    handled = true;
                    break;
                case btnDelete:
                    if ( 0 <= sel && Confirm( altServerDelete ) ) {
                        ServerDelete( serverSets[ sel ].path );
                        DrawServerForm();
                    }
                    else if ( sel < 0 )
                        FrmAlert( altSelectSet );
                    handled = true;
                    break;
            }
            break;
        }

        default:
            break;
    }
    return handled;
}



static void UpdateFTPDisplay( void )
{
    CtlSetLabel( GetObjectPtr( trgPassword ),
         pref.server.havePassword ? "assigned" : "prompt" );
}


static void SaveFields( void )
{
    CopyFromField( pref.server.host, fldServer, MAX_SERVER_SIZE + 1 );
    CopyFromField( pref.server.user, fldUser, MAX_USER_SIZE + 1 );
    CopyFromField( pref.server.baseDir, fldBaseDir, MAX_BASEDIR_SIZE + 1 );
    pref.server.port = StrAToI( FldGetTextPtr( GetObjectPtr( fldPort ) ) );
    if ( pref.server.port == 0 )
        pref.server.port = 21;

    SavePrefs();
}



Boolean Handler_frmFTP( EventType* event )
{
    Boolean handled;

    handled = false;

    switch (event->eType) {
        case frmOpenEvent: {
            Char portString[ 6 ];

            if ( pref.server.port == 0 ) {
                pref.server.port = 21;
                SavePrefs();
            }

            StrPrintF( portString, "%u", pref.server.port );
            SetField( fldPort, portString );

            SetField( fldServer, pref.server.host );
            SetField( fldUser, pref.server.user );
            SetField( fldBaseDir, pref.server.baseDir );
            SetValue( chkExportAfterBackup, pref.exportAfterBackup );

            UpdateFTPDisplay();
            FrmDrawForm( FrmGetActiveForm() );

            handled = true;
            break;
        }

        case frmCloseEvent: {
            SaveFields();

            handled = false;
            break;
        }

        case ctlSelectEvent:
            switch( event->data.ctlEnter.controlID ) {
                case btnOK:
                    SaveFields();
                    FrmReturnToForm( 0 );
                    handled = true;
                    break;
                case trgPassword: {
                    Char password[ MAX_PASSWORD_SIZE + PASSWORD_SCRATCH_SPACE ];
                    password[ 0 ] = 0;
                    switch ( FrmCustomResponseAlert( altPassword, " your server",
                                 "  Leave blank if not needed.",
                                 " but sent unencrypted to the server",
                                 password, MAX_PASSWORD_SIZE, NULL ) ) {
                        case 0:
                            MemSet( pref.server.password, MAX_PASSWORD_SIZE + PASSWORD_SCRATCH_SPACE, 0 );
                            PasswordEncrypt( pref.server.password, password );
                            MemSet( password, MAX_PASSWORD_SIZE + PASSWORD_SCRATCH_SPACE, 0 );
                            pref.server.havePassword = true;
                            break;
                        case 1:
                            pref.server.havePassword = 0;
                            MemSet( pref.server.password, MAX_PASSWORD_SIZE + PASSWORD_SCRATCH_SPACE, 0 );
                            break;
                        default:
                            break;
                    }
                    SavePrefs();
                    UpdateFTPDisplay();
                    handled = true;
                    break;
                }
                case chkExportAfterBackup:
                    if ( pref.server.havePassword ) {
                        pref.exportAfterBackup = GetValue( chkExportAfterBackup );
                        SavePrefs();
                    }
                    else {
                        SetValue( chkExportAfterBackup, false );
                        pref.exportAfterBackup = false;
                        SavePrefs();
                        FrmAlert( altNeedPassword );
                    }
                    handled = true;
                    break;
            }
            break;

        default:
            break;
    }
    return handled;
}




