/* $Id: zip.c,v 1.21 2007/02/11 15:19:29 arpruss Exp $ */
//#define DEBUG
// uses globals! (useDmitrysTrick, init, zlib)

#include <PalmOS.h>
#include <VFSMgr.h>
#include "Backup.h"
#include "crc32.h"

#define ZIP_BUFFER_SIZE 60000

#ifdef DEBUG
# define DebugMessage( s ) Message( IMPORTANT, s )
#else
# define DebugMessage( s )
#endif

#define FLAG_FTEXT    ( 1 << 0 )
#define FLAG_FHCRC    ( 1 << 1 )
#define FLAG_FEXTRA   ( 1 << 2 )
#define FLAG_FNAME    ( 1 << 3 )
#define FLAG_FCOMMENT ( 1 << 4 )

static Boolean init = false;


ZIP_SECTION static Err WriteCryptBlock( ZFile* z );
ZIP_SECTION static Err MyFileSize( ZFile* z, UInt32* sizeP, Boolean reposition );
ZIP_SECTION static Err MyWrite( ZFile* z, UInt32 size, void* data, UInt32* didP );
ZIP_SECTION static Err MyRead( ZFile* z, UInt32 size, void* data, UInt32* didP );
ZIP_SECTION static Err MyTell( ZFile* z, UInt32* offset );
ZIP_SECTION static Err MySeek( ZFile* z, UInt16 origin, Int32 offset );
ZIP_SECTION static Err MyClose( ZFile* z );
ZIP_SECTION static Err Copy( z_streamp strm );
ZIP_SECTION static Err MyDeflate(z_streamp strm, int flush, Boolean compress);
ZIP_SECTION static Err MyInflate(z_streamp strm, int flush, Boolean compress);
ZIP_SECTION static void CryptReset( ZFile* zref );
ZIP_SECTION static Err ZVFSWriteFlush( ZFile* zref );
ZIP_SECTION static Err ZVFSReadFill( ZFile* zref, Boolean* eofP );

static Err WriteCryptBlock( ZFile* z )
{
    UInt32 did;
    Err    err;
    UInt8  block[ MAX_BLOCK_LENGTH ];

    z->crypt.Crypt( 1, block, z->cryptBlock, z->crypt.key, true );

    err = VFSFileWrite( z->ref, z->crypt.blockSize, block, &did );

    if ( err != errNone )
        return err;

    if ( did != z->crypt.blockSize )
        return 1;

    z->totInCryptBlock = z->crypt.blockSize;
    z->posInCryptBlock = 0;

    return errNone;
}


static Err MyWrite( ZFile* z, UInt32 size, void* data, UInt32* didP )
{
    Err err;
    Boolean lowMem;

    if ( z->crypt.Crypt == NULL )
        return VFSFileWrite( z->ref, size, data, didP );

    if ( didP != NULL )
        *didP = 0;

    err = errNone;
    lowMem = false;

    while ( err == errNone && 0 < size ) {
        if ( ! lowMem && z->posInCryptBlock == 0 && 2 * z->crypt.blockSize <= size ) {
            UInt32 numBlocks;
            UInt32 did;
            UInt32 rounded;
            UInt32 curSize;
            void*  buffer;

            if ( 60000 < size )
                curSize = 60000;
            else
                curSize = size;

            numBlocks = curSize / z->crypt.blockSize;
            rounded   = numBlocks * z->crypt.blockSize;

            buffer = MemPtrNew( rounded );

            if ( buffer == NULL ) {
                lowMem = true;
                continue;
            }

            z->crypt.Crypt( numBlocks, buffer, data, z->crypt.key, true );

            did = 0;
            err = VFSFileWrite( z->ref, rounded, buffer, &did );

            MemPtrFree( buffer );

            if ( err != errNone )
                return err;

            if ( did != rounded )
                return 1;

            if ( didP != NULL )
                *didP += rounded;

            size -= rounded;
            data  = ( UInt8* )data + rounded;
        }
        else if ( z->posInCryptBlock < z->crypt.blockSize ) {
            UInt32 toMove;

            if ( size < z->crypt.blockSize - z->posInCryptBlock ) {
                toMove = size;
            }
            else {
                toMove = z->crypt.blockSize - z->posInCryptBlock;
            }

            MemMove( z->cryptBlock + z->posInCryptBlock, data, toMove );

            if ( didP != NULL )
                *didP += toMove;

            z->posInCryptBlock += toMove;
            data = ( UInt8* )data + toMove;
            size -= toMove;
        }
        else {
            err = WriteCryptBlock( z );
            /* TODO: optimize for larger sizes */
        }
    }

    return err;
}


static Err ReadCryptBlock( ZFile* z )
{
    UInt32 did;
    Err    err;
    UInt8  block[ MAX_BLOCK_LENGTH ];

    err = VFSFileRead( z->ref, z->crypt.blockSize, block, &did );

    if ( err != errNone ) {
        return err;
    }

    if ( did % z->crypt.blockSize != 0 ) {
        return 1;
    }

    if ( did == 0 )
        return vfsErrFileEOF;

    z->crypt.Crypt( 1, z->cryptBlock, block, z->crypt.key, false );
//    { Char s[55]; StrPrintF(s, "%x %x", z->cryptBlock[ 0 ], block[ 0 ] ); Message(INFO,s); }

    if ( VFSFileEOF( z->ref ) ) /* last block ends with the size */ {
        z->totInCryptBlock = z->cryptBlock[ z->crypt.blockSize - 1 ];
    }
    else {
        z->totInCryptBlock = z->crypt.blockSize;
    }
    
    z->posInCryptBlock = 0;
    return errNone;
}


static Err MyRead( ZFile* z, UInt32 size, void* data, UInt32* didP )
{
    Err err;

    if ( z->crypt.Crypt == NULL )
        return VFSFileRead( z->ref, size, data, didP );

    if ( didP != NULL )
        *didP = 0;

    err = errNone;

    while ( err == errNone && 0 < size ) {
        if ( z->posInCryptBlock < z->totInCryptBlock ) {
            UInt32 toMove;

            if ( size < z->totInCryptBlock - z->posInCryptBlock ) {
                toMove = size;
            }
            else {
                toMove = z->totInCryptBlock  - z->posInCryptBlock;
            }

            MemMove( data, z->cryptBlock + z->posInCryptBlock, toMove );

            if ( didP != NULL )
                *didP += toMove;

            z->posInCryptBlock += toMove;
            size               -= toMove;
            data = ( UInt8* )data + toMove;
        }
        else if ( z->crypt.blockSize <= size ) {
            UInt32 numBlocks;
            UInt32 did;
            UInt32 rounded;

            numBlocks = size / z->crypt.blockSize;
            rounded   = numBlocks * z->crypt.blockSize;

            err = VFSFileRead( z->ref, rounded, data, &did );

            if ( err != errNone && err != vfsErrFileEOF )
                return err;

            if ( did == 0 )
                return vfsErrFileEOF;

            if ( did % z->crypt.blockSize != 0 )
                return 1;

            z->crypt.Crypt( numBlocks, data, data, z->crypt.key, false );

            if ( VFSFileEOF( z->ref ) ) {
                did -= z->crypt.blockSize - ( ( UInt8* )data )[ did - 1 ];
            }

            size -= did;
            data  = ( UInt8* )data + did;
            if ( didP != NULL )
                *didP += did;
        }
        else {
            err = ReadCryptBlock( z );
            /* TODO: optimize for larger sizes */
        }
    }

    return err;
}


static Err MyTell( ZFile* z, UInt32* offsetP )
{
    Err    err;

    err = VFSFileTell( z->ref, offsetP );

    if ( z->crypt.Crypt == NULL || err != errNone )
        return err;

    if ( z->write )
        *offsetP += z->posInCryptBlock;
    else
        *offsetP -= z->totInCryptBlock - z->posInCryptBlock;

    return errNone;
}

/* Only for read-only streams */
static Err MySeek( ZFile* z, UInt16 origin, Int32 offset )
{
    Err err;
    Int32 rounded;

    if ( z->crypt.Crypt == NULL )
        return VFSFileSeek( z->ref, origin, offset );

    if ( z->write )
        return 1;

    if ( origin == vfsOriginEnd ) {
        Err err;
        UInt32 fileSize;

        err = MyFileSize( z, &fileSize, false );

        if ( err != errNone )
            return err;

        offset += ( Int32 )fileSize;
    }
    else if ( origin == vfsOriginCurrent ) {
        Err   err;
        Int32 cur;

        err = VFSFileTell( z->ref, &cur );

        if ( err != errNone )
            return err;
            
        if ( VFSFileEOF( z->ref ) )
            cur -= z->crypt.blockSize - z->posInCryptBlock;
        else
            cur -= z->totInCryptBlock - z->posInCryptBlock;

        offset += cur;
    }

    rounded = offset / z->crypt.blockSize * z->crypt.blockSize;

    err = VFSFileSeek( z->ref, vfsOriginBeginning, rounded );

    if ( err != errNone )
        return err;

    if ( rounded == offset ) {
        CryptReset( z );
        return errNone;
    }

    err = ReadCryptBlock( z );

    if ( err != errNone )
        return err;

    z->posInCryptBlock = offset - rounded;
    
    if ( z->totInCryptBlock < z->posInCryptBlock )
        return 1;
        
    return errNone;
}


/* Only for read-only streams */
static Err MyFileSize( ZFile* z, UInt32* sizeP, Boolean reposition )
{
    UInt32 roundSize;
    UInt32 curPos;
    UInt8  inBlock[ MAX_BLOCK_LENGTH ];
    UInt8  outBlock[ MAX_BLOCK_LENGTH ];
    Err    err;

    if ( z->crypt.Crypt == NULL )
        return VFSFileSize( z->ref, sizeP );

    if ( z->write )
        return 1;

    err = VFSFileSize( z->ref, &roundSize );

    if ( err != errNone )
        return err;

    if ( roundSize % z->crypt.blockSize != 0 )
        return 1;

    if ( roundSize < z->crypt.blockSize ) {
        *sizeP = 0;

        return errNone;
    }

    if ( reposition ) {
        err = VFSFileTell( z->ref, &curPos );
    
        if ( err != errNone )
            return err;
    }

    roundSize -= z->crypt.blockSize;

    err = VFSFileSeek( z->ref, vfsOriginBeginning, roundSize );

    if ( err != errNone )
        return err;

    err = VFSFileRead( z->ref, z->crypt.blockSize, inBlock, NULL );

    if ( err != errNone )
        return err;

    z->crypt.Crypt( 1, outBlock, inBlock, z->crypt.key, z->write );

    *sizeP = roundSize + outBlock[ z->crypt.blockSize - 1 ];

    if ( reposition ) {
        err = VFSFileSeek( z->ref, vfsOriginBeginning, curPos );
    }

    return err;
}


static Err MyClose( ZFile* z )
{
    Err   err;

    if ( z->crypt.Crypt == NULL || ! z->write ) {
        MemZero( z->crypt.key, MAX_KEY_LENGTH );
        return VFSFileClose( z->ref );
    }

    /* TODO: Dump last block. */

    if ( z->posInCryptBlock == z->crypt.blockSize ) {
        err = WriteCryptBlock( z );

        if ( err != errNone ) {
            VFSFileClose( z->ref );
            return err;
        }
    }

    z->cryptBlock[ z->crypt.blockSize - 1 ] = z->posInCryptBlock;

    err = WriteCryptBlock( z );

    VFSFileClose( z->ref );

    return err;
}


static Err Copy( z_streamp strm )
{
    UInt16 size;

    if ( strm->avail_in == 0 )
        return Z_STREAM_END;
    if ( strm->avail_out == 0 )
        return Z_OK;
    size = strm->avail_in;
    if ( strm->avail_out < size )
        size = strm->avail_out;
    MemMove( strm->next_out, strm->next_in, size );
    strm->next_in += size;
    strm->next_out += size;
    strm->avail_in -= size;
    strm->avail_out -= size;

    if ( strm->avail_in == 0 )
        return Z_STREAM_END;
    return Z_OK;
}


static Err MyDeflate(z_streamp strm, int flush, Boolean compress)
{
    if ( compress )
        return deflateSZ( 0, strm, flush );
    else
        return Copy( strm );

}

static Err MyInflate(z_streamp strm, int flush, Boolean compress)
{
    if ( compress )
        return inflateSZ( 0, strm, flush );
    else
        return Copy( strm );
}

void CleanupZip( void )
{
    if ( init ) {
        zlib_close( 0, NULL );
        CloseCRC();
        init = false;
    }
}



void InitZip( void )
{
    if ( ! init ) {
        InitCRC();
        if ( errNone == zlib_open( 0 ) ) {
            init = true;
        }
        else {
            CloseCRC();
            FatalError( altErrorZLib );
        }
    }
}



UInt32 GetUInt32LE( UInt8* p )
{
     UInt32 x;

     p+= 3;

     x = *p--;
     x = x * 256 + *p--;
     x = x * 256 + *p--;
     x = x * 256 + *p--;

     return x;
}



UInt16 ConvertUInt16LE( UInt16 v )
{
     UInt8* p;

     p = ( UInt8* )&v ;

     return p[ 0 ] + 256 * p[ 1 ];
}



void PutUInt32LE( UInt8* p, UInt32 x )
{
    *p++ = x & 0xFF;
    x >>= 8;
    *p++ = x & 0xFF;
    x >>= 8;
    *p++ = x & 0xFF;
    x >>= 8;
    *p = x;
}


UInt32 ToUnixDate( UInt32 palmDate )
{
    /* This does NOT adjust for timezone */
    if ( palmDate == 0 )
        return 0;

    return palmDate - 2082844800ul;
}


UInt32 FromUnixDate( UInt32 unixDate )
{
    /* This does NOT adjust for timezone */
    if ( unixDate == 0 )
        return 0;

    return unixDate + 2082844800ul;
}


static void CryptReset( ZFile* zref )
{
    zref->posInCryptBlock = 0;
    zref->totInCryptBlock = 0;
    MemZero( zref->cryptBlock, MAX_BLOCK_LENGTH );
}


Err ZVFSOpenWrite( UInt16 vol, const Char* path, ZFile* zref, Int16 level, UInt32 mdate, CryptInfo* cryptP )
{
    Err err;
//    Char myName[ MAX_FILENAME ];
    UInt8 header[ 10 ];
    UInt16 i;

    MemSet( zref, sizeof( ZFile ), 0 );

    zref->write       = true;

    if ( cryptP != NULL && cryptP->Crypt != NULL ) {
        zref->crypt = *cryptP;
        CryptReset( zref );
    }
    else
        zref->crypt.Crypt = NULL;

#if 0
    if ( 0 <= level && ( StrLen( path ) < 3 || StrCompare( path + StrLen( path ) - 3, ".gz" ) ) ) {
        StrPrintF( myName, "%s.gz", path );
    }
    else {
        StrCopy( myName, path );
    }
#endif

    err = VFSFileOpen( vol, path, vfsModeWrite, &( zref->ref ) );
    
    zref->compressed = ( 0 <= level );

    if ( err != errNone ) {
        return err;
    }

    if ( mdate != 0 ) {
        VFSFileSetDate( zref->ref, vfsFileDateCreated, mdate );
        VFSFileSetDate( zref->ref, vfsFileDateModified, mdate );
        VFSFileSetDate( zref->ref, vfsFileDateAccessed, mdate );
    }
    zref->date = mdate;

    zref->buffer = MemPtrNew( ZIP_BUFFER_SIZE );

    if ( zref->buffer == NULL )
        return 1;

    zref->bufferSize = ZIP_BUFFER_SIZE;

    zref->date       = mdate;

    if ( ! zref->compressed ) {
        zref->trailerSize = 0;
        zref->headerSize  = 0;

        return errNone;
    }
    else {
        zref->trailerSize = 8;
        zref->headerSize  = 10;
    }

    i = 0;
    header[ i++ ] = 31;  // MAGIC 1
    header[ i++ ] = 139; // MAGIC 2
    header[ i++ ] = 8;   // deflate
    header[ i++ ] = 0;   // FLG

    PutUInt32LE( header + i, ToUnixDate( mdate ) );
    i += 4;

    header[ i++ ] = 0;   // XFL
    header[ i++ ] = 255; // OS

    err = MyWrite( zref, zref->headerSize, header, NULL );

    if ( err != errNone ) {
        MemPtrFree( zref->buffer );
        MyClose( zref );
        return err;
    }

    MemSet( &( zref->zs ), sizeof( zref->zs ), 0 );

    err = deflateInit2_SZ( 0, &( zref->zs ), level, Z_DEFLATED, -13, 6, Z_DEFAULT_STRATEGY, ZLIB_VERSION,
        sizeof( zref->zs ) );

    if ( err != Z_OK ) {
        MemPtrFree( zref->buffer );
        MyClose( zref );
        return err;
    }

    zref->zs.avail_in = 0;
    zref->zs.next_out  = zref->buffer + zref->inBuffer;
    zref->zs.avail_out = zref->bufferSize - zref->inBuffer;

    zref->crc32 = 0;

    return err;
}



Int32 SkipPastNul( ZFile* z )
{
    UInt8 c;
    Int32 count;
    Err   err;

    count = 0;

    do {
        UInt32 didRead;

        err = MyRead( z, 1, &c, &didRead );

        if ( err != errNone ) {
            return -1;
        }

        count++;
    } while ( c != 0 );

    return count;
}



Err ZVFSOpenRead( UInt16 vol, const Char* path, ZFile* zref, CryptInfo* cryptP )
{
    Err err;
    UInt8 header[ 10 ];
    Boolean hiddenVolume;
    static  Boolean useDmitrysTrick = false;

    hiddenVolume = ! pref.testMode && ! StrNCompare( path, "/PALM_DM/", 9 );

    MemSet( zref, sizeof( ZFile ), 0 );

    zref->compressed = ( 3 < StrLen( path ) && ! StrCompare( path + StrLen( path ) - 3, ".gz" ) );
    zref->write      = false;

    if ( cryptP != NULL && cryptP->Crypt != NULL ) {
        zref->crypt = *cryptP;
        CryptReset( zref );
    }
    else
        zref->crypt.Crypt = NULL;

    if ( hiddenVolume ) {
        err = errNone;
        if ( ! useDmitrysTrick )
            err = VFSFileOpen( vol, path, vfsModeRead, &( zref->ref ) );
        if ( useDmitrysTrick || err != errNone ) {
            err = VFSFileOpen( vol, path, 0, &( zref->ref ) );
            if ( err == errNone )
                useDmitrysTrick = true;
        }
    }
    else {
        err = VFSFileOpen( vol, path, vfsModeRead, &( zref->ref ) );
    }


    if ( err != errNone )
        return err;

    MemSet( &( zref->zs ), sizeof( zref->zs ), 0 );

    err = MyFileSize( zref, &( zref->rawSize ), true );

    if ( err != errNone ) {
        MyClose( zref );
        return 1;
    }

    if ( ! zref->compressed ) {
        zref->trailerSize = 0;
        zref->headerSize  = 0;
    }
    else {
        zref->trailerSize = 8;
        zref->headerSize  = 10;
    }

    /* skip gzip header */
    if ( zref->compressed ) {
        UInt32 didRead;
        UInt8  flag;

        err = MyRead( zref, zref->headerSize, header, &didRead );

        if ( err != errNone || header[ 0 ] != 31 || header[ 1 ] != 139 ) {
            MyClose( zref );
            return 1;
        }

        zref->date = FromUnixDate( GetUInt32LE( header + 4 ) );

        flag = header[ 3 ];

        if ( flag & FLAG_FEXTRA ) {
            UInt16 size;

            err = MyRead( zref, 2, &size, &didRead );

            if ( err != errNone ) {
                MyClose( zref );
                return 1;
            }

            zref->headerSize += 2;

            size = ConvertUInt16LE( size );

            zref->headerSize += size;

            err = MySeek( zref, vfsOriginBeginning, zref->headerSize );

            if ( err != errNone ) {
                MyClose( zref );
                return 1;
            }
        }

        if ( flag & FLAG_FNAME ) {
            Int32 size;

            size = SkipPastNul( zref );

            if ( size < 0 ) {
                MyClose( zref );
                return 1;
            }

            zref->headerSize += size;
        }

        if ( flag & FLAG_FCOMMENT ) {
            Int32 size;

            size = SkipPastNul( zref );

            if ( size < 0 ) {
                MyClose( zref );
                return 1;
            }

            zref->headerSize += size;
        }

        if ( flag & FLAG_FHCRC ) {
            UInt16 hcrc;

            err = MyRead( zref, 2, &hcrc, &didRead );

            if ( err != errNone ) {
                MyClose( zref );
                return 1;
            }
            
            zref->headerSize += 2;
        }
        
        zref->rawRead = zref->headerSize;
    }
    else {
        zref->rawRead = 0;
        if ( errNone != VFSFileGetDate( zref->ref, vfsFileDateModified, &( zref->date ) ) )
            zref->date = 0;
    }

    zref->buffer = MemPtrNew( ZIP_BUFFER_SIZE );

    if ( zref->buffer == NULL )
        return 1;

    zref->bufferSize = ZIP_BUFFER_SIZE;

    if ( zref->compressed ) {
        err = inflateInit2_SZ( 0, &( zref->zs ), -15, ZLIB_VERSION, sizeof( zref->zs ) );

        if ( err != Z_OK ) {
            MemPtrFree( zref->buffer );
            MyClose( zref );
            return err;
        }
    }

    zref->crc32 = 0;

    return err;
}


static Err ZVFSWriteFlush( ZFile* zref )
{
    Err err;

    if ( zref->inBuffer == 0 )
        return errNone;

    err = MyWrite( zref, zref->inBuffer, zref->buffer, NULL );

    zref->inBuffer = 0;

    return err;
}


static Err ZVFSReadFill( ZFile* zref, Boolean* eofP )
{
    Err err;
    UInt32 didRead;
    UInt32 left;
    UInt16 doRead;

    if ( zref->rawSize <= zref->trailerSize + zref->rawRead ) {

        if ( eofP != NULL )
            *eofP = true;

        return errNone;
    }

    doRead = zref->bufferSize;

    left = zref->rawSize - zref->trailerSize - zref->rawRead;

    if ( left < doRead )
        doRead = left;

    err = MyRead( zref, doRead, zref->buffer, &didRead );

    zref->bufferPos = 0;
    zref->inBuffer  = didRead;
    zref->rawRead  += didRead;

    if ( eofP != NULL )
        *eofP = ( err == vfsErrFileEOF || doRead < zref->bufferSize );

    if ( err != errNone && err != vfsErrFileEOF )
        return err;

    return errNone;
}



Err ZVFSWrite( ZFile* zref, UInt16 length, UInt8* s )
{
    Err err;

    if ( length == 0 )
        return errNone;

    zref->totalUncompressedSize += length;

    if ( zref->compressed ) {
        zref->crc32 = CRC32( zref->crc32, s, length );
    }

    zref->zs.avail_in = length;
    zref->zs.next_in  = s;

    while ( 0 < zref->zs.avail_in ) {
        if ( zref->bufferSize <= zref->inBuffer ) {
            err = ZVFSWriteFlush( zref );

            if ( err != errNone )
                return err;
        }

        zref->zs.next_out    = zref->buffer + zref->inBuffer;
        zref->zs.avail_out   = zref->bufferSize - zref->inBuffer;

        err = MyDeflate( &( zref->zs ), Z_NO_FLUSH, zref->compressed );

        zref->flushed = ( err == Z_STREAM_END );

        if ( err != Z_OK && err != Z_STREAM_END )
            return err;

        zref->inBuffer = zref->zs.next_out - zref->buffer;
    }

    return errNone;
}



Err ZVFSRead( ZFile* zref, UInt16 length, UInt8* s, UInt16* didReadP )
{
    Err     err;
    UInt16  didRead;
    Boolean eof;

    if ( length == 0 )
        return errNone;

    zref->zs.avail_out = length;
    zref->zs.next_out  = s;

    err = Z_OK;
    eof = zref->noMore;

    while ( 0 < zref->zs.avail_out && err == Z_OK && ( ! eof || zref->bufferPos < zref->inBuffer ) ) {
                                                                                                       
        if ( zref->inBuffer <= zref->bufferPos ) {
            Err err;

            err = ZVFSReadFill( zref, &eof );

            if ( err != errNone ) {
                if ( didReadP != NULL )
                    *didReadP = zref->zs.next_out - s;

                return err;
            }
            else if ( zref->inBuffer <= zref->bufferPos ) {
//                err = Z_STREAM_END;
                eof = true;
            }
        }

        zref->zs.avail_in = zref->inBuffer - zref->bufferPos;
        zref->zs.next_in  = zref->buffer   + zref->bufferPos;

        err = MyInflate( &( zref->zs ), Z_SYNC_FLUSH, zref->compressed );

        if ( err == Z_OK || err == Z_STREAM_END || ( err == Z_BUF_ERROR && eof ) ) {
            zref->bufferPos = zref->zs.next_in - zref->buffer;
            if ( err == Z_BUF_ERROR ) {
                err = Z_STREAM_END;
            }
        }
    }

    didRead = zref->zs.next_out - s;

    if ( eof && ( err == Z_STREAM_END || didRead < length ) ) {
        zref->noMore = true;
    }

    if ( didReadP != NULL )
        *didReadP = didRead;
//{ Char s[55]; StrPrintF( s, "<%x %d>", didRead, didRead ); Message(s); }

    if ( ( err == Z_OK || err == Z_STREAM_END ) && zref->compressed ) {
        zref->crc32 = CRC32( zref->crc32, s, didRead );
    }

    if ( err == Z_STREAM_END ) {
//Message("\b!");
        if ( eof && didRead < length )
            return vfsErrFileEOF;
        else
            return errNone;
    }

    return err;
}



// We assume that VFS errors and zlib errors do not overlap
Err ZVFSCloseWrite( ZFile* zref )
{
    Err   err;
    UInt8 trailer[ 8 ];

    if ( zref->flushed )
        err = Z_STREAM_END;
    else
        err = Z_OK;

    while ( err != Z_STREAM_END ) {
        if ( zref->bufferSize <= zref->inBuffer ) {
            err = ZVFSWriteFlush( zref );
            if ( err != errNone )
                break;
        }

        zref->zs.next_out  = zref->buffer + zref->inBuffer;
        zref->zs.avail_out = ( zref->bufferSize - zref->inBuffer );
        zref->zs.next_in   = NULL;
        zref->zs.avail_in  = 0;

        err = MyDeflate( &( zref->zs ), Z_FINISH, zref->compressed );

        zref->inBuffer = zref->zs.next_out - zref->buffer;

        if ( err != Z_STREAM_END && err != Z_OK ) {
            break;
        }
    }

    if ( err == Z_STREAM_END ) {
        err = ZVFSWriteFlush( zref );
    }
    else {
        ZVFSWriteFlush( zref );
    }

    if ( zref->compressed ) {
        PutUInt32LE( trailer, zref->crc32 );
        PutUInt32LE( trailer + 4, zref->totalUncompressedSize );

        if ( err == errNone ) {
            err = MyWrite( zref, 8, trailer, NULL );
        }
        else {
            MyWrite( zref, 8, trailer, NULL );
        }

        deflateEndSZ( 0, &( zref->zs ) );
    }
    MemPtrFree( zref->buffer );
    if ( zref->date != 0 ) {
        VFSFileSetDate( zref->ref, vfsFileDateCreated, zref->date );
        VFSFileSetDate( zref->ref, vfsFileDateModified, zref->date );
        VFSFileSetDate( zref->ref, vfsFileDateAccessed, zref->date );
    }
    MyClose( zref );

    return err;
}




Err ZVFSCloseRead( ZFile* zref, Boolean check )
{
     Err err;

     err = errNone;

     if ( zref->compressed ) {
         inflateEndSZ( 0, &( zref->zs ) );

         if ( zref->rawSize <= zref->rawRead + zref->trailerSize ) {
             err = MySeek( zref, vfsOriginEnd, -8 );

             if ( err == errNone ) {
                 UInt8 crc[ 4 ];

                 err = MyRead( zref, 4, crc, NULL );

                 if ( err != errNone || GetUInt32LE( crc ) != zref->crc32 ) {
                     err = 2;
                 }
             }
         }
         else {
             err = 3;   
         }
     }
     MemPtrFree( zref->buffer );
     MyClose( zref );

     return err;
}



Err ZVFSFileClose( ZFile* zref, Boolean check )
{
     Err err;

     if ( zref->write )
         err = ZVFSCloseWrite( zref );
     else
         err = ZVFSCloseRead( zref, check );
         
     return err;
}



Err ZVFSFileSize( ZFile* zref, UInt32* sizeP )
{
     Err err;
     UInt32 lastPos;
     UInt32 didRead;
     UInt8  length[ 4 ];

     if ( zref->write )
         return 1;

     *sizeP = zref->rawSize;

     if ( ! zref->compressed )
         return errNone;

     if ( *sizeP < 18 )
         return 1;

     err = MyTell( zref, &lastPos );

     if ( err != errNone )
         return err;

     err = MySeek( zref, vfsOriginEnd, -4 );

     if ( err != errNone )
         return err;

     err = MyRead( zref, 4, length, &didRead );

     if ( ( err == errNone || err == vfsErrFileEOF ) && didRead == 4 ) {
         *sizeP = GetUInt32LE( length );
         err = errNone;
     }
     else {
         err = 1;
     }

     if ( err == errNone )
         err = MySeek( zref, vfsOriginBeginning, lastPos );
     else
         MySeek( zref, vfsOriginBeginning, lastPos );

     return err;
}



#if 0
Err ZVFSRewind( ZFile* zref )
{
    Err err;

    if ( zref->write )
        return 1; // not supported

    if ( zref->compressed )
        inflateEndSZ( 0, &( zref->zs ) );

    err = VFSFileSeek( zref->ref, vfsOriginBeginning, zref->headerSize );

    zref->inBuffer  = 0;
    zref->bufferPos = 0;
    zref->noMore    = 0;

    if ( err != errNone || ! zref->compressed )
        return err;

    zref->rawRead = zref->headerSize;
    zref->crc32 = 0;

    MemSet( &( zref->zs ), sizeof( zref->zs ), 0 );

    return inflateInit2_SZ( 0, &( zref->zs ), -15, ZLIB_VERSION, sizeof( zref->zs ) );
}
#endif


/* Only for readonly uncompressed files. */
Err ZCRC32( ZFile* zref, UInt32* crcP )
{
    Err err;
    UInt32 size = zref->rawSize;

    err = MySeek( zref, vfsOriginBeginning, 0 );

    if ( err != errNone )
        return err;

    while ( 0 < size ) {
        UInt32 size1;

        size1 = zref->bufferSize;

        if ( size < size1 )
            size1 = size;

        err = MyRead( zref, size1, zref->buffer, NULL );

        if ( err != errNone )
            break;

        *crcP = CRC32( *crcP, zref->buffer, size1 );
        
        size -= size1;
    }

    return err;
}



// closes file iff crc32 != NULL 
Err ZVFSStats( ZFile* zref, UInt32* date, UInt32* size, UInt32* crc32 )
{
     Err err;
//     UInt32 lastPos;

//     err = VFSFileTell( zref->ref, &lastPos );

//     if ( err != errNone )
//         return err;

     if ( zref->write )
         return 1;

     err = errNone;

     if ( date != NULL )
         *date = zref->date;

     if ( zref->compressed && ( size != NULL || crc32 != NULL ) ) {
         UInt8 trailer[ 8 ];

         err = MySeek( zref, vfsOriginEnd, -8 );

         if ( err == errNone )
             err = MyRead( zref, 8, trailer, NULL );

         if ( err == errNone ) {
             if ( size != NULL )
                 *size  = GetUInt32LE( trailer + 4 );
             if ( crc32 != NULL ) {
                 *crc32 = GetUInt32LE( trailer );
                 inflateEndSZ( 0, &( zref->zs ) );
             }
         }
     }
     else {
         if ( size != NULL )
             *size = zref->rawSize;

         if ( crc32 != NULL ) {
             *crc32 = 0;

             err = ZCRC32( zref, crc32 );
         }
     }

     if ( crc32 != NULL ) {
         MemPtrFree( zref->buffer );
         MyClose( zref );
     }

     return err;
}

